#include <openssl/evp.h>
#include <openssl/err.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <xcipher.h>

#if OPENSSL_VERSION_NUMBER < 0x10100000L
#define EVP_MD_CTX_new   EVP_MD_CTX_create
#define EVP_MD_CTX_free  EVP_MD_CTX_destroy
#endif


#define XCIPHER_EVP_ERROR_HANDLER() \
        rc = -1; goto XCIPHER_CATCH_ERROR
       
#define XCIPHER_PROPERTIES_LEN_ALGORITHM    128
#define XCIPHER_PROPERTIES_LEN_KEY          128
#define XCIPHER_PROPERTIES_LEN_IV           32

/* 
  XCIPHER_io_t in.properties  
  ο ˰ ߰  ׸ ߰ 
*/
struct XCIPHER_properties_s
{
    char algorithm[XCIPHER_PROPERTIES_LEN_ALGORITHM];
    int input_type;  // reserved
    int output_type; // reserved
    int key_len;
    unsigned char key[XCIPHER_PROPERTIES_LEN_KEY];
    int iv_len;
    unsigned char iv[XCIPHER_PROPERTIES_LEN_IV];
};

typedef struct XCIPHER_properties_s XCIPHER_properties_t;


/* local function prototype */
static const EVP_CIPHER * get_evp_cipher(XCIPHER_io_t *io);
static const EVP_MD * get_evp_md(XCIPHER_io_t *io);
static int aes_cbc_encrypt(XCIPHER_io_t *io); // aes/cbc ȣȭ
static int aes_cbc_decrypt(XCIPHER_io_t *io); // aes/cbc ȣȭ
static int base64_encode(XCIPHER_io_t *io);   // base64 encode
static int base64_decode(XCIPHER_io_t *io);   // base64 decode
static int digest_message(XCIPHER_io_t *io);  // md5, sha1, sha2, sha3 ܹ ȣȭ (ȣȭ  )
static int hex_encode(XCIPHER_io_t *io);      // hexa encode
static int hex_decode(XCIPHER_io_t *io);      // hexa decode
static int input_validate(XCIPHER_io_t *io);
static int openssl_get_err_msg(char *msg, const char *user_msg);
static int parse_algorithm(XCIPHER_io_t *io);
static void string_to_lowercase(char *dest, const char *src);

/* --------------------------------------- function body ---------------------------------------------------------------------- */
int XCIPHER_set_io(XCIPHER_io_t *io, const char *in_data, int in_data_size, char *out_buf, int out_buf_size)
{
    int rc = 0;
    
    if((rc = XCIPHER_set_input(io, in_data, in_data_size, XCIPHER_DATA_TYPE_BUFFER)) == -1) {
        return rc;
    }
    if((rc = XCIPHER_set_output(io, out_buf, out_buf_size, XCIPHER_DATA_TYPE_BUFFER)) == -1) {
        return rc;
    }
    return rc;
}

int XCIPHER_set_input(XCIPHER_io_t *io, const char *in_data, int in_data_size, int input_type)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;

    if(io == NULL) {
        return -1;
    }
    if(in_data == NULL) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "input data is null.");
        return -1;
    }
    if(in_data_size < 0) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "input data size error.");
        return -1;
    }

    io->in.data = in_data;
    io->in.size = in_data_size;
    properties->input_type = input_type;

    return 0;
}

int XCIPHER_set_output(XCIPHER_io_t *io, char *out_buf, int out_buf_size, int output_type)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;

    if(io == NULL) {
        return -1;
    }

    if(out_buf == NULL) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "output buffer is null.");
        return -1;
    }

    if(out_buf_size < 0) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "output buffer size error.");
        return -1;
    }

    io->out.data = out_buf;
    io->out.size = out_buf_size;
    properties->output_type = output_type;

    return 0;
}

const char *XCIPHER_get_algorithm(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->algorithm;
}

int XCIPHER_get_input_type(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->input_type;
}

int XCIPHER_get_output_type(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->output_type;
}

const char *XCIPHER_get_key(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->key;
}

int XCIPHER_get_key_len(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->key_len;
}

const char *XCIPHER_get_iv(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->iv;
}

int XCIPHER_get_iv_len(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    return properties->iv_len;
}

void XCIPHER_set_algorithm(XCIPHER_io_t *io, const char *algorithm)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    strncpy(properties->algorithm, algorithm, XCIPHER_PROPERTIES_LEN_ALGORITHM);
}

void XCIPHER_set_key(XCIPHER_io_t *io, const char *key, int key_len)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    memcpy(properties->key, key, key_len);
    properties->key_len = key_len;
}

void XCIPHER_set_iv(XCIPHER_io_t *io, const char *iv, int iv_len)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
   
    memcpy(properties->iv, iv, iv_len);
    properties->iv_len = iv_len;
}


/*********************************************************************************************************************************
 * FUNCTION  : XCIPHER_encrypt
 * FUNCTION   : ȣȭ ó
 ********************************************************************************************************************************/
int XCIPHER_encrypt(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
    int rc = 0;
    char algorithm[128];
   
    string_to_lowercase(algorithm, properties->algorithm);

    if((rc = input_validate(io)) != 0) {
        return rc;
    }
    if(get_evp_cipher(io) != NULL) {
        rc = aes_cbc_encrypt(io);
    } else if(get_evp_md(io) != NULL) {
        rc = digest_message(io);
    } else if(strncmp(algorithm, "base64", 6) == 0) {
        rc = base64_encode(io);
    } else if(strncmp(algorithm, "hex", 3) == 0) {
        rc = hex_encode(io);
    } else {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s  ʴ ȣȭ ˰Դϴ.", algorithm);
        return -1;
    }
    return rc;
}

/*********************************************************************************************************************************
 * FUNCTION  : XCIPHER_decrypt
 * FUNCTION   : ȣȭó 
 ********************************************************************************************************************************/
int XCIPHER_decrypt(XCIPHER_io_t *io)
{
    XCIPHER_properties_t *properties = (XCIPHER_properties_t *)io->in.properties;
    int rc = 0;
    char algorithm[128];
   
    string_to_lowercase(algorithm, properties->algorithm);

    if((rc = input_validate(io)) != 0) {
        return rc;
    }

    if(get_evp_cipher(io) != NULL) {
        rc = aes_cbc_decrypt(io);
    } else if(strncmp(algorithm, "base64", 6) == 0) {
        rc = base64_decode(io);
    } else if(strncmp(algorithm, "hex", 3) == 0) {
        rc = hex_decode(io);
    } else {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s  ʴ ȣȭ ˰̰ų ȣȭ  ˰Դϴ.", algorithm);
        return -1;
    }
    return rc;
}

/*********************************************************************************************************************************
 * FUNCTION  : input_validate
 * FUNCTION   : INPUT Validation check
 ********************************************************************************************************************************/
static int input_validate(XCIPHER_io_t *io)
{
    if(io->in.data == NULL) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "ó Ͱ(XCIPHER_io.in.data)  ʾҽϴ.");
        return -1;
    }
    if(io->in.size <= 0) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "ó  (XCIPHER_io.in.size) ߸ Ǿϴ.");
        return -1;
    }
    if(io->out.data == NULL) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "  output buffer(XCIPHER_io.out.data)  ʾҽϴ.");
        return -1;
    }
    if(io->out.size <= 0) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "  output buffer ũ(XCIPHER_io.out.size) 0 Ǵ -  ʾҽϴ.");
        return -1;
    }

    return 0;
}

/*********************************************************************************************************************************
 * FUNCTION  : get_evp_md
 * FUNCTION   : md5, sha1, sha2, sha3 迭 ˰  EVP_MD return
 ********************************************************************************************************************************/
static const EVP_MD * get_evp_md(XCIPHER_io_t *io)
{
    const char   *algorithm = XCIPHER_get_algorithm(io);

    if(strcasecmp(algorithm, "md5") == 0) {
        return EVP_md5();
    } else if(strcasecmp(algorithm, "sha1") == 0) {
        return EVP_sha1();
    } else if(strcasecmp(algorithm, "sha224") == 0) {
        return EVP_sha224();
    } else if(strcasecmp(algorithm, "sha256") == 0) {
        return EVP_sha256();
    } else if(strcasecmp(algorithm, "sha512") == 0) {
        return EVP_sha512();
    } else if(strcasecmp(algorithm, "sha512/224") == 0) {
        return EVP_sha512_224();
    } else if(strcasecmp(algorithm, "sha512/256") == 0) {
        return EVP_sha512_256();
    } else if(strcasecmp(algorithm, "sha3/224") == 0 || strcasecmp(algorithm, "sha3_224") == 0) {
        return EVP_sha3_224();
    } else if(strcasecmp(algorithm, "sha3/256") == 0 || strcasecmp(algorithm, "sha3_256") == 0) {
        return EVP_sha3_256();
    } else if(strcasecmp(algorithm, "sha3/512") == 0 || strcasecmp(algorithm, "sha3_512") == 0) {
        return EVP_sha3_512();
    }
    return NULL;
}
/*********************************************************************************************************************************
 * FUNCTION  : digest_message
 * FUNCTION   : ַ binary data   1Ʈ ͸ 2Ʈ hexa ڵ ȯϴ Լ
 ********************************************************************************************************************************/
static int digest_message(XCIPHER_io_t *io)
{
    int   rc = 0;
    const unsigned char *message = (const unsigned char *)io->in.data;
    unsigned char digest[2048]; 
    unsigned int  digest_len;
	EVP_MD_CTX   *mdctx = NULL;
    const EVP_MD *evp_md = NULL;
    const char   *algorithm = XCIPHER_get_algorithm(io);

    evp_md = get_evp_md(io);

    if(evp_md == NULL) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "%s  ʴ ˰Դϴ.", algorithm);
        XCIPHER_EVP_ERROR_HANDLER();
    }
    
	if((mdctx = EVP_MD_CTX_new()) == NULL) {
        openssl_get_err_msg(io->out.err_msg, "EVP_MD_CTX_new() ȣ : ");
        XCIPHER_EVP_ERROR_HANDLER();
    }

	if(EVP_DigestInit_ex(mdctx, evp_md, NULL) != 1) {
        openssl_get_err_msg(io->out.err_msg, "EVP_DigestInit_ex() ȣ : ");
        XCIPHER_EVP_ERROR_HANDLER();
    }

	if(EVP_DigestUpdate(mdctx, message, io->in.size) != 1) {
        openssl_get_err_msg(io->out.err_msg, "EVP_DigestUpdate() ȣ : ");
        XCIPHER_EVP_ERROR_HANDLER();
    }

	if(EVP_DigestFinal_ex(mdctx, digest, &digest_len) != 1) {
        openssl_get_err_msg(io->out.err_msg, "EVP_DigestFinal_ex() ȣ : ");
        XCIPHER_EVP_ERROR_HANDLER();
    }

    if(digest_len > io->out.size) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "output buffer ũⰡ ۽ϴ.[ũ(%d) > outputũ(%d)].", digest_len, io->out.size);
        XCIPHER_EVP_ERROR_HANDLER();
    }
    
    io->out.size = digest_len;
    memcpy(io->out.data, digest, digest_len);
    
XCIPHER_CATCH_ERROR:
    if(mdctx != NULL) {
        EVP_MD_CTX_free(mdctx);
    }

    return rc;
}

/*********************************************************************************************************************************
 * FUNCTION  : hex_encode
 * FUNCTION   : ַ binary data   1Ʈ ͸ 2Ʈ hexa ڵ ȯϴ Լ
 ********************************************************************************************************************************/
static int hex_encode(XCIPHER_io_t *io)
{
    int     rc = 0;
    char    tmp_buf[io->in.size * 2 + 1];
    int     is_lowercase = 0;
    int     idx;
    unsigned char *in_data = (unsigned char *)io->in.data;

    if(io->in.size * 2 > io->out.size) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "output buffer ũⰡ  ʽϴ.(INPUT ũ 2谡 ʿմϴ.)");
        return -1;
    }
    if(strncmp(XCIPHER_get_algorithm(io), "hex", 3) == 0) { // algorithm Ī ҹ̸ 16 ҹڷ
        is_lowercase = 1;
    }
    memset(tmp_buf, 0x00, sizeof(tmp_buf));

    for(idx = 0; idx < io->in.size; idx++) {
        if(is_lowercase) {
            snprintf(&tmp_buf[idx * 2], 3, "%.02x", (int)in_data[idx]);
        } else {
            snprintf(&tmp_buf[idx * 2], 3, "%.02X", (int)in_data[idx]);
        }
    }

    io->out.size = io->in.size * 2;
    memcpy(io->out.data, tmp_buf, io->out.size);

    return rc;
}

/*********************************************************************************************************************************
 * FUNCTION  : hex_decode
 * FUNCTION   : 2Ʈ  hex ڵ 1Ʈ  binary ȯϴ Լ
 ********************************************************************************************************************************/
static int hex_decode(XCIPHER_io_t *io)
{
    int rc = 0;
    int idx, out_size = 0;
    unsigned char tmp_src[io->in.size + 1];
    unsigned char tmp_dst[io->in.size];
    unsigned char ch1, ch2;

    if(io->in.size > io->out.size * 2) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "Output buffer ũⰡ  ʽϴ.(Input ũ 1/2 ̻ ʿմϴ.)");
        return -1;
    }
    if(io->in.size % 2 == 1) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "ȯ input data ̰ ¦ ƴմϴ.");
        return -1;
    }

    memset(tmp_src, 0x00, sizeof(tmp_src));
    memset(tmp_dst, 0x00, sizeof(tmp_dst));
    string_to_lowercase(tmp_src, io->in.data);
   
    for(idx = 0; idx < io->in.size; idx++) {
        if('0' <= tmp_src[idx] && tmp_src[idx] <= '9') {
            ch1 = tmp_src[idx] - '0';
        } else if('a' <= tmp_src[idx] && tmp_src[idx] <= 'f') {
            ch1 = tmp_src[idx] - 'a' + 10;
        } else {
            snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "Input data hex ƴ Ͱ ԵǾ ֽϴ.");
            return -1;
        }
        idx++;
        if('0' <= tmp_src[idx] && tmp_src[idx] <= '9') {
            ch2 = tmp_src[idx] - '0';
        } else if('a' <= tmp_src[idx] && tmp_src[idx] <= 'f') {
            ch2 = tmp_src[idx] - 'a' + 10;
        } else {
            snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "Input data hex ƴ Ͱ ԵǾ ֽϴ.");
            return -1;
        }
        tmp_dst[out_size++] = (ch1 << 4) | ch2;
    }
    memcpy(io->out.data, tmp_dst, out_size);
    io->out.size = out_size;

    return rc;
}


/*********************************************************************************************************************************
 * FUNCTION  : aes_cbc_encrypt
 * FUNCTION   : AES128, AES192, AES256  ȣȭ ó
 ********************************************************************************************************************************/
static int aes_cbc_encrypt(XCIPHER_io_t *io)
{
    int rc = 0;
    int out_len = 0, howmany = 0, block_size;
    const EVP_CIPHER *cipher;
    EVP_CIPHER_CTX *ectx = NULL;
    int iv_len = 0, key_len = 0;
    char iv[XCIPHER_PROPERTIES_LEN_IV];
    char key[XCIPHER_PROPERTIES_LEN_KEY];
    char out_buf[(io->in.size < io->out.size ? io->out.size : io->in.size) + 256];
    int padding_size = 0;
    unsigned char padding_buf[256];
    int idx = 0;

    memset(key, 0x00, sizeof(key));
    memset(iv, 0x00, sizeof(iv));
    memset(out_buf, 0x00, sizeof(out_buf));

    key_len = XCIPHER_get_key_len(io);
    iv_len = XCIPHER_get_iv_len(io);

    if(key_len <= 0) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "ȣȭ key  ʾҽϴ.");
        XCIPHER_EVP_ERROR_HANDLER();       
    }
    if(iv_len <= 0) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "iv  ʾҽϴ.");
        XCIPHER_EVP_ERROR_HANDLER();       
    }

    memcpy(key, XCIPHER_get_key(io), key_len);
    memcpy(iv, XCIPHER_get_iv(io), iv_len);

    ectx = EVP_CIPHER_CTX_new();

    if((cipher = get_evp_cipher(io)) == NULL) {
        XCIPHER_EVP_ERROR_HANDLER();
    }
    if(EVP_EncryptInit_ex(ectx, cipher, NULL, key, iv) != 1) {
        openssl_get_err_msg(io->out.err_msg, "Encypt Init Error");
        XCIPHER_EVP_ERROR_HANDLER();
    }
    if(EVP_EncryptUpdate(ectx, out_buf, &howmany, io->in.data, io->in.size) != 1) {
        openssl_get_err_msg(io->out.err_msg, "Data encrypt processing error");
        XCIPHER_EVP_ERROR_HANDLER();
    }
    out_len = howmany;

    if(EVP_EncryptFinal_ex (ectx, out_buf + out_len, &howmany) != 1){
        XCIPHER_EVP_ERROR_HANDLER();
    }
    out_len += howmany;

    if(out_len > io->out.size) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "Ҵ output buffer size ۽ϴ.");
        XCIPHER_EVP_ERROR_HANDLER();
    }
    memcpy(io->out.data, out_buf, out_len);
    io->out.size = out_len;

XCIPHER_CATCH_ERROR:
    if(ectx != NULL) {
        EVP_CIPHER_CTX_free(ectx);
    }

    return rc;
}

/*********************************************************************************************************************************
 * FUNCTION  : aes_cbc_decrypt
 * FUNCTION   : AES128, AES192, AES256  ȣȭ ó
 ********************************************************************************************************************************/
static int aes_cbc_decrypt(XCIPHER_io_t *io)
{
    int rc = 0;
    int out_len = 0, howmany = 0, block_size;
    const EVP_CIPHER *cipher;
    EVP_CIPHER_CTX *ectx = NULL;
    int iv_len = 0, key_len = 0;
    char iv[XCIPHER_PROPERTIES_LEN_IV];
    char key[XCIPHER_PROPERTIES_LEN_KEY];
    char out_buf[(io->in.size < io->out.size ? io->out.size : io->in.size) + 256];
    int padding_size = 0;
    unsigned char padding_buf[256];
    int idx = 0;

    memset(key, 0x00, sizeof(key));
    memset(iv, 0x00, sizeof(iv));
    memset(out_buf, 0x00, sizeof(out_buf));

    key_len = XCIPHER_get_key_len(io);
    iv_len = XCIPHER_get_iv_len(io);

    if(key_len <= 0) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "ȣȭ key  ʾҽϴ.");
        XCIPHER_EVP_ERROR_HANDLER();   
    }
    if(iv_len <= 0) {
        snprintf(io->out.err_msg, sizeof(io->out.err_msg), "%s", "iv  ʾҽϴ.");
        XCIPHER_EVP_ERROR_HANDLER();
    }

    memcpy(key, XCIPHER_get_key(io), key_len);
    memcpy(iv, XCIPHER_get_iv(io), iv_len);

    ectx = EVP_CIPHER_CTX_new();

    if((cipher = get_evp_cipher(io)) == NULL) {
        XCIPHER_EVP_ERROR_HANDLER();
    }

    if(EVP_DecryptInit_ex(ectx, cipher, NULL, key, iv) != 1) {
        openssl_get_err_msg(io->out.err_msg, "Encypt Init Error");
        XCIPHER_EVP_ERROR_HANDLER();
    }

    if(EVP_DecryptUpdate(ectx, out_buf, &howmany, io->in.data, io->in.size) != 1) {
        openssl_get_err_msg(io->out.err_msg, "Data encrypt processing error");
        XCIPHER_EVP_ERROR_HANDLER();
    }
    out_len = howmany;

    if(EVP_DecryptFinal_ex (ectx, out_buf + out_len, &howmany) != 1){
        openssl_get_err_msg(io->out.err_msg, "DecryptFinal error");
        XCIPHER_EVP_ERROR_HANDLER();
    }
    out_len += howmany;
    if(out_len > io->out.size) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "Ҵ output buffer size ۽ϴ.");
        XCIPHER_EVP_ERROR_HANDLER();
    }
    memcpy(io->out.data, out_buf, out_len);
    io->out.size = out_len;

XCIPHER_CATCH_ERROR:
    if(ectx != NULL) {
        EVP_CIPHER_CTX_free(ectx);
    }

    return rc;
}

/*********************************************************************************************************************************
 * FUNCTION  : base64_encode
 * FUNCTION   : BASE64 encoding ó
 ********************************************************************************************************************************/
static int base64_encode(XCIPHER_io_t *io)
{
    unsigned char out_buf[(io->in.size * 4 / 3 + 4) < io->out.size ? io->out.size : (io->in.size * 4 / 3 + 4)];
    unsigned char *input = (unsigned char *)io->in.data;
    int in_data_size = (int)io->in.size;
    int out_len = 0;

    int ret = EVP_EncodeBlock(out_buf, input, in_data_size);
    out_len = strlen(out_buf);

    if(out_len > io->out.size) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "Ҵ output buffer size ۽ϴ.");
        return -1;
    }
    strcpy(io->out.data, out_buf);
    io->out.size = out_len;

    return 0;
}

/*********************************************************************************************************************************
 * FUNCTION  : base64_decode
 * FUNCTION   : BASE64 decoding ó
 ********************************************************************************************************************************/
static int base64_decode(XCIPHER_io_t *io)
{
    unsigned char out_buf[io->in.size];
    unsigned char *input = (unsigned char *)io->in.data;
    int in_data_size = (int)io->in.size;
    int pad_count = 0;
    int ret = 0;

    if (input[io->in.size - 1] == '=') { pad_count++; }
    if (input[io->in.size - 2] == '=') { pad_count++; }
    memset(out_buf, 0x00, sizeof(out_buf));
    if((ret = EVP_DecodeBlock(out_buf, input, in_data_size)) < 0) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "Base64 decoding  ߻Ͽϴ.");
        return -1;
    }

    ret -= pad_count;
    if(ret > io->out.size) {
        snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "Ҵ output buffer size ۽ϴ.");
        return -1;
    }
    memcpy(io->out.data, out_buf, ret);
    io->out.size = ret;

    return 0;
}

/*********************************************************************************************************************************
 * FUNCTION  : openssl_get_err_msg
 * FUNCTION   : EVP  ڵ忡  ڿ   ޽ 
 ********************************************************************************************************************************/
static int openssl_get_err_msg(char *msg, const char *user_msg)
{
    int last_error = 0, error = 0;
    char tmp_err_msg[128];
    while((last_error = ERR_get_error())) {
        error = last_error;
    }
    if(error != 0) {
        ERR_error_string(error, tmp_err_msg);
        snprintf(msg, XCIPHER_OUT_LEN_ERR_MSG, "%s: %s", user_msg, tmp_err_msg);
        return error;
    } else {
        snprintf(msg, XCIPHER_OUT_LEN_ERR_MSG, "%s: Unkown error.", user_msg);
    }
    return error;
}

/*********************************************************************************************************************************
 * FUNCTION  : string_to_lowercase
 * FUNCTION   : ڿ ҹڷ .
 ********************************************************************************************************************************/
static void string_to_lowercase(char *dest, const char *src)
{
    do {
        *(dest++) = tolower(*src);
    } while(*(src++));
}

/*********************************************************************************************************************************
 * FUNCTION  : get_evp_cipher
 * FUNCTION   : algorithm ´ EVP_CIPHER return Ѵ.
 ********************************************************************************************************************************/
static const EVP_CIPHER * get_evp_cipher(XCIPHER_io_t *io)
{
    const char *algorithm = XCIPHER_get_algorithm(io);
   
    if(strncasecmp(algorithm, "aes256/cbc", 10) == 0) {
        return EVP_aes_256_cbc();
    } else if(strncasecmp(algorithm, "aes192/cbc", 10) == 0) {
        return EVP_aes_192_cbc();
    } else if(strncasecmp(algorithm, "aes128/cbc", 10) == 0) {
        return EVP_aes_128_cbc();
    }
   
    snprintf(io->out.err_msg, XCIPHER_OUT_LEN_ERR_MSG, "%s  ʴ /ȣȭ ˰Դϴ.", algorithm);
    return NULL;
}
